/*********************************************************
 *
 *    Xdemo -- defunc demonstration example
 *             for X Window System(TM) 
 *
 *    Author: Ke Jin, Queen's U. Canada 
 *
 *    This program  can be included in other programs
 *    or software products, provided the source codes
 *    are also supplied.
 *
 *    to compile: cc Xdemo.c -ldfc -lm -lX11 -o Xdemo
 *
 *    Note:  The X Window System is a Trademark of MIT.
 *
 *********************************************************/

#include <stdio.h>
#include <string.h> 
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <math.h>
#include <defunc.h>

#define  hinttext "Input from the parent window"

#define rint(x)  ceil(x)


#if NeedFunctionPrototypes
  void XDrawFunc(
  Display *pDisplay, 
  Window  win,
  GC      gc,
  double  (*fp)(),
  int     left, 
  int     up,
  int     width,
  int     height,
  double  x0,
  double  y0,
  double  x1,
  double  y1,
  double  dx)
#else
  XDrawFunc(pDisplay, win, gc, fp, left, up, width, height, 
                                   x0, y0, x1, y1, dx)
  Display *pDisplay;
  Window  win;
  GC      gc;
  double  (*fp)();
  int     left, up;        /* left-up corner of the area        */
  int     width, height;   /* width and height of the area      */
  double  x0, y0, x1, y1;
  double  dx;        
#endif
{
    double x, a, b;
    int    X1, Y1, X2, Y2;

    if(fp==0) return ;

    a = width/(x1-x0);
    b = height/(y0-y1);

    for(x=x0;x<=x1;x=x+dx)
    {
        X1 = left + rint(a*(x-x0));
        Y1 = up   + rint(b*(fp(x)-y1));
	X2 = left + rint(a*(x+dx-x0));
        Y2 = up   + rint(b*(fp(x+dx)-y1));
        if(Y1>=up&&
	   Y1<=height+up&&
	   Y2>=up&&
	   Y2<=height+up&&
	   abs(Y1-Y2)<=height)
	{
	    XDrawLine(pDisplay, win, gc, X1, Y1, X2, Y2);
        }
    }
};

#if NeedFunctionPrototypes
  int main(int argc, char** argv)
#else
  int main(argc, argv)
  int argc;
  char** argv;
#endif
{
    double     (*fp)()=0;
    double     (*tempfp)()=0;
    char       expr[128];
    char       tempexpr[128];
    double     x0, y0, x1, y1, a, b;
    char       sx0[32], sx1[32], sy0[32], sy1[32];

    int      i, winleft, winup, pid;
    unsigned border_width = 20;
    unsigned border = 3;
    unsigned depth = 0;
    unsigned long whitepix, blackpix;
    unsigned long winwidth=400, winheight=300;
    unsigned long arealeft=rint(0.10*winwidth);
    unsigned long areaup  =rint(0.15*winheight);
    unsigned long areawidth  = winwidth -2*arealeft;
    unsigned long areaheight = winheight-2*areaup;

    Display    *pDisplay;
    Screen     *pScreen;
    Window     win, rootwin;
    GC         gc; 
    XEvent     event;
    XSizeHints sizehint;

    pDisplay = XOpenDisplay("");
    if(pDisplay==0)
    {
	perror("Open display");
	exit(1);
    }

    pScreen = XDefaultScreenOfDisplay(pDisplay);
    if(pScreen==0)
    {
	perror("Get default screen");
        exit(1);
    }

    blackpix = XBlackPixelOfScreen(pScreen);
    whitepix = XWhitePixelOfScreen(pScreen);

    sizehint.width = winwidth;
    sizehint.height = winheight;
    sizehint.flags = PPosition; 

    rootwin = DefaultRootWindow(pDisplay);
    win = XCreateSimpleWindow(pDisplay, rootwin,
           sizehint.x, sizehint.y, sizehint.width, sizehint.height,
	   border_width, border, whitepix); 

/*HPUX Fix: arg #8 */
    XSetStandardProperties(pDisplay,
			   win, 
			   "Xdemo (defunc 1.3)",  /* window name */
			   "Xdemo",        /* icon   name */
			   None, argv, argc, &sizehint);

    gc = XCreateGC(pDisplay, win, 0, 0);
    XSetBackground(pDisplay, gc, whitepix);
    XSetForeground(pDisplay, gc, blackpix);
    XSetLineAttributes(pDisplay, gc, 
		       1,          /* line width */
		       LineSolid,  /* line style */
		       CapNotLast, /* cap  style */
		       JoinMiter); /* join style */

    XSelectInput(pDisplay, win, ExposureMask);
    XMapWindow(pDisplay, win);

    matha2z();

    XNextEvent(pDisplay, &event);    /* flush window */
    XClearWindow(pDisplay, win);
    XDrawRectangle(pDisplay, win, gc, arealeft, areaup,
                  winwidth-2*arealeft, winheight-2*areaup);
    XDrawImageString(pDisplay, win, gc,
		rint(winwidth/2.)-3*strlen(hinttext), rint(winheight/2.),
	        hinttext, strlen(hinttext));

    XSendEvent(pDisplay, win, 0, ExposureMask, &event);
  /*  XCheckWindowEvent(pDisplay, &event);*/ /* flush window */

/*HPUX X11R6 requires 4 params ... */
     	XCheckWindowEvent(pDisplay,win,ExposureMask, &event);  /* flush window */
/****/

    printf("\n\tEnter an defunc expression. e.g.:\n\n");
    printf("\t\tsin(x)*cos(1/x)\n");
    printf("\t\tsin(x)/abs(x)\n");
    printf("\t\t[0, 3*pi][0, 10] x*exp(-x*sin(x)^2)\n");
    printf("\t\t[,3*sqrt(pi)][-.5,.5] exp(-x)*sin(x*x)\n");
    printf("\t\t[-20,20][0,1] abs(sin(sin(x)/x))\n");
    printf("\n\t(\"quit\" or \"exit\" to exit).\n");

    dfopen("[-2*pi, 2*pi][-1.2,1.2]");

    for(;;)
    {
	printf("\n\tinput > ");
        fgets(tempexpr, 127, stdin);

        XClearWindow(pDisplay, win);

	for(i=0;i<strlen(tempexpr);i++)
	{
	    if(tempexpr[i]=='\n'||tempexpr[i]==';') tempexpr[i]='\0';
        }

        if(strncmp(tempexpr, "quit", 4)==0||
           strncmp(tempexpr, "exit", 4)==0)
        {
	    printf("\n\t\t--------- Bye! ---------\n\n");
	    break;
        }

	tempfp = dfopen(tempexpr);   

        x0 = dfcmin(1);
        x1 = dfcmax(1);
        y0 = dfcmin(2);
	y1 = dfcmax(2);

	if(x0==x1) 
	{
	    sprintf(sx0, "%.3g", x0);
	    x0 = x0-1.0;
	    x1 = x1+1.0;
            XDrawImageString(pDisplay, win, gc,
		  arealeft+rint(areawidth/2.-3*strlen(sx0)),
		  areaup+areaheight+15,
		  sx0, strlen(sx0));
        }

        if(y0==y1) 
	{
	    sprintf(sy0, "%.3g", y0);
	    y0 = y0-1.0;
	    y1 = y1+1.0;
            XDrawImageString(pDisplay, win, gc,
		  arealeft-8*strlen(sy0), 
		  areaup+rint(areaheight/2.),
		  sy0, strlen(sy0));
        }

	sprintf(sx0, "%.3g", x0);
        sprintf(sx1, "%.3g", x1);
	sprintf(sy0, "%.3g", y0);
	sprintf(sy1, "%.3g", y1);

	if(tempfp==0)
	{
	    printf("\t\t%s\n", dferror());
        }
	else
	{
	    dfclose(fp);
	    fp = tempfp;
	    strncpy(expr, tempexpr, strlen(tempexpr)+1);
        }

	XGetGeometry(pDisplay, win, &rootwin, 
              &winleft, &winup, &winwidth, &winheight, 
	      &border_width, &depth);
	arealeft   = rint(0.1*winwidth);
	areaup     = rint(0.15*winheight);
        areawidth  = winwidth -2*arealeft;
        areaheight = winheight-2*areaup;

	a = areawidth/(x1-x0);
	b = areaheight/(y0-y1);

	if(y0*y1<0)
	  /* draw y axial */
	{
	    XDrawLine(pDisplay, win, gc,
	    arealeft, areaup+rint(b*(0.0-y1)),
	    arealeft+areawidth, areaup+rint(b*(0.0-y1)));
        }

	if(x0*x1<0)
	  /* draw x axial */
	{
            XDrawLine(pDisplay, win, gc,
	    arealeft+rint(a*(0.0-x0)), areaup,
	    arealeft+rint(a*(0.0-x0)), areaup+areaheight);
        }

	XDrawRectangle(pDisplay, win, gc,
	      arealeft, areaup, areawidth, areaheight);

        XDrawImageString(pDisplay, win, gc,
	      arealeft-10, areaup+areaheight+15, 
	      sx0, strlen(sx0));
        XDrawImageString(pDisplay, win ,gc,
	      arealeft+areawidth-10, areaup+areaheight+15, 
	      sx1, strlen(sx1));
        XDrawImageString(pDisplay, win ,gc,
	      arealeft-6*strlen(sy1)-3, areaup+10, 
	      sy1, strlen(sy1));
        XDrawImageString(pDisplay, win ,gc,
	      arealeft-6*strlen(sy0)-3, areaup+areaheight, 
	      sy0, strlen(sy0));

        if(fp!=0)
	  /* draw function curve, function body */
	{
	    XDrawImageString(pDisplay, win, gc,
	          arealeft, areaup-10, 
		  getexprbody(expr), strlen(getexprbody(expr)));

            XSetLineAttributes(pDisplay, gc,
		  2,          /* line width */
	          LineSolid,  /* line style */
                  CapNotLast, /* cap  style */
                  JoinMiter); /* join style */
	    XDrawFunc(pDisplay, win, gc, fp, 
		  arealeft, areaup, areawidth, areaheight, 
		  x0, y0, x1, y1, 1/a);
            XSetLineAttributes(pDisplay, gc,
		  1,          /* line width */
                  LineSolid,  /* line style */
		  CapNotLast, /* cap  style */
		  JoinMiter); /* join style */
        }

        XSendEvent(pDisplay, win, 0, ExposureMask, &event);

/*HPUX X11R6 requires 4 params ... */
     	XCheckWindowEvent(pDisplay,win,ExposureMask, &event);  /* flush window */
/****/
        printf("\n\t\tcurrent arguments are %s, %s\n",
	      getarguname(1), getarguname(2));
    }

    XUnmapWindow(pDisplay, win);
    XFreeGC(pDisplay, gc);
    XDestroyWindow(pDisplay, win);
    XCloseDisplay(pDisplay);

    return 0;
};
