/*************************************************************
 *
 *    Copyright (c) 1993 1994  by Ke Jin, Queen's U. Canada. 
 *
 *    Permission to use, copy, modify, and distribute 
 *    this software and its documentation without fee
 *    is granted, provided that the author's name and
 *    this copyright notice are retained.
 *
 *    This software can be included in other programs
 *    or software products, provided the source codes
 *    are also supplied without charge.
 * 
 * -----------------------------------------------------------
 *
 *    defunc.c -- defunc high level module 
 *
 *    public  function : dfopen();
 *                       dfclose();
 *                       dfcloseall();
 *
 *    private function : getfreehdl();
 *                       tdfopen();
 *
 *                       hdl00() to hdl31();
 *
 *    private variable : tr_arr[];
 *                     : hdl_arr[];
 *
 ***********************************************************/

#include <stdio.h>
#include <malloc.h>
#include <ctype.h> 
#include <string.h>

#include "dfctree.h"
#include "dfcsymtable.h"
#include "defunc.h"

#ifdef __cplusplus   /* for c++ */
  extern "C" {
#endif

#define MAXHANDLE 32

typedef Node* Tree;
typedef double (*FP)();

#if NeedFunctionPrototypes
  static Tree   tr_arr[MAXHANDLE];  /* predeclare static member */
  static FP     hdl_arr[MAXHANDLE]; /* predeclare static member */
#else
  extern Tree   tr_arr[MAXHANDLE];  /* predeclare static member */
  extern FP     hdl_arr[MAXHANDLE]; /* predeclare static member */
#endif /* predeclare private function */

#if NeedFunctionPrototypes
  static int getfreehdl(void)
#else
  static int getfreehdl() 
#endif
/* return the index of free handle on success. return -1 on fail */
{
    int i;

    for(i=0;i<MAXHANDLE;i++)
    {
	if(tr_arr[i]==(Node*)0) return i;
    }

    return -1;    /* if no free handle */
};

#if NeedFunctionPrototypes
  static double (*tdfopen(char* expression))(double x, double y)
#else
  static double (*tdfopen(expression))() 
  char* expression;
#endif
/* Return handle function or NULL */
{
    int i, size;

    if((i=getfreehdl())==-1)  return 0; /* no free handle */
    
    size = exparse(expression);
    if(size<=0) return 0;

    tr_arr[i] = (Node*)malloc(sizeof(Node)*size); 
    if(tr_arr[i]==0) 
    {
	perror("alloc memory for new parse tree");
	exit(1);
    }

    getparsetree(tr_arr[i]);
    dfcreduce(tr_arr[i]);

    return hdl_arr[i];
};

#if NeedFunctionPrototypes
  double (*dfopen(char* expr))(double x, double y)
#else
  double (*dfopen(expr))()
  char* expr;
#endif 
/* Accept full expression as well as expression 
 * body.  Return handle function or NULL */
{
    double (*fnctptr)();

    char* exprname;

    fnctptr = tdfopen(expr);
    exprname = gettreename();

    if(exprname!=0)
    {
	if(isfnctdecl()) 
	{
	    namefnct(exprname, fnctptr);
        }
	else if(fnctptr!=0)
	{
	    namecnst(exprname, fnctptr());
        }
    }

    return fnctptr;
};

#if NeedFunctionPrototypes
  int dfclose(double (*fnctptr)())
#else
  int dfclose(fnctptr)
  double (*fnctptr)();
#endif
/* discarding a dynamic function(getten from dfopen) if it is 
 * not on the global external token table. On success, the 
 * function  return 0. On fail return -1 */
{
    int i;
    if(getfnctname(fnctptr)!=0) return -1; 
    /* function still on symbol table can't be closed */ 

    for(i=0;i<MAXHANDLE;i++)
    {
	if(fnctptr==hdl_arr[i])
	{
	    if(tr_arr[i]!=0) free(tr_arr[i]);
	    tr_arr[i]=0;
	    return 0; /* success */
        }
    }

    return -1; /* fail to close */
};

#if NeedFunctionPrototypes
  int dfcloseall(void)
#else
  int dfcloseall()
#endif
/* close all dynamic functions which are not on token table. 
 * Return the total number of handler freed by this calling.  */
{
     int i, j;

     for(i=0, j=0; i<MAXHANDLE; i++)
     {
	 if(tr_arr[i]!=0)  /* unfreed handle */ 
	 {
	     if(dfclose(hdl_arr[i])==0) /* success */
	     {
	       tr_arr[i]=0;
	       j++;
             }
         }
     }

     return j;
};

/* ------------------------- private members  ------------------------ */
static  Tree   tr_arr[MAXHANDLE];

#if NeedFunctionPrototypes
 static double hdl00(double x,double y){return evaluate(tr_arr[ 0],x,y);};
 static double hdl01(double x,double y){return evaluate(tr_arr[ 1],x,y);};
 static double hdl02(double x,double y){return evaluate(tr_arr[ 2],x,y);};
 static double hdl03(double x,double y){return evaluate(tr_arr[ 3],x,y);};
 static double hdl04(double x,double y){return evaluate(tr_arr[ 4],x,y);};
 static double hdl05(double x,double y){return evaluate(tr_arr[ 5],x,y);};
 static double hdl06(double x,double y){return evaluate(tr_arr[ 6],x,y);};
 static double hdl07(double x,double y){return evaluate(tr_arr[ 7],x,y);};
 static double hdl08(double x,double y){return evaluate(tr_arr[ 8],x,y);};
 static double hdl09(double x,double y){return evaluate(tr_arr[ 9],x,y);};
 static double hdl10(double x,double y){return evaluate(tr_arr[10],x,y);};
 static double hdl11(double x,double y){return evaluate(tr_arr[11],x,y);};
 static double hdl12(double x,double y){return evaluate(tr_arr[12],x,y);};
 static double hdl13(double x,double y){return evaluate(tr_arr[13],x,y);};
 static double hdl14(double x,double y){return evaluate(tr_arr[14],x,y);};
 static double hdl15(double x,double y){return evaluate(tr_arr[15],x,y);};
 static double hdl16(double x,double y){return evaluate(tr_arr[16],x,y);};
 static double hdl17(double x,double y){return evaluate(tr_arr[17],x,y);};
 static double hdl18(double x,double y){return evaluate(tr_arr[18],x,y);};
 static double hdl19(double x,double y){return evaluate(tr_arr[19],x,y);};
 static double hdl20(double x,double y){return evaluate(tr_arr[20],x,y);};
 static double hdl21(double x,double y){return evaluate(tr_arr[21],x,y);};
 static double hdl22(double x,double y){return evaluate(tr_arr[22],x,y);};
 static double hdl23(double x,double y){return evaluate(tr_arr[23],x,y);};
 static double hdl24(double x,double y){return evaluate(tr_arr[24],x,y);};
 static double hdl25(double x,double y){return evaluate(tr_arr[25],x,y);};
 static double hdl26(double x,double y){return evaluate(tr_arr[26],x,y);};
 static double hdl27(double x,double y){return evaluate(tr_arr[27],x,y);};
 static double hdl28(double x,double y){return evaluate(tr_arr[28],x,y);};
 static double hdl29(double x,double y){return evaluate(tr_arr[29],x,y);};
 static double hdl30(double x,double y){return evaluate(tr_arr[30],x,y);};
 static double hdl31(double x,double y){return evaluate(tr_arr[31],x,y);};
#else
 static double hdl00(x,y) double x,y; {return evaluate(tr_arr[ 0],x,y);};
 static double hdl01(x,y) double x,y; {return evaluate(tr_arr[ 1],x,y);};
 static double hdl02(x,y) double x,y; {return evaluate(tr_arr[ 2],x,y);};
 static double hdl03(x,y) double x,y; {return evaluate(tr_arr[ 3],x,y);};
 static double hdl04(x,y) double x,y; {return evaluate(tr_arr[ 4],x,y);};
 static double hdl05(x,y) double x,y; {return evaluate(tr_arr[ 5],x,y);};
 static double hdl06(x,y) double x,y; {return evaluate(tr_arr[ 6],x,y);};
 static double hdl07(x,y) double x,y; {return evaluate(tr_arr[ 7],x,y);};
 static double hdl08(x,y) double x,y; {return evaluate(tr_arr[ 8],x,y);};
 static double hdl09(x,y) double x,y; {return evaluate(tr_arr[ 9],x,y);};
 static double hdl10(x,y) double x,y; {return evaluate(tr_arr[10],x,y);};
 static double hdl11(x,y) double x,y; {return evaluate(tr_arr[11],x,y);};
 static double hdl12(x,y) double x,y; {return evaluate(tr_arr[12],x,y);};
 static double hdl13(x,y) double x,y; {return evaluate(tr_arr[13],x,y);};
 static double hdl14(x,y) double x,y; {return evaluate(tr_arr[14],x,y);};
 static double hdl15(x,y) double x,y; {return evaluate(tr_arr[15],x,y);};
 static double hdl16(x,y) double x,y; {return evaluate(tr_arr[16],x,y);};
 static double hdl17(x,y) double x,y; {return evaluate(tr_arr[17],x,y);};
 static double hdl18(x,y) double x,y; {return evaluate(tr_arr[18],x,y);};
 static double hdl19(x,y) double x,y; {return evaluate(tr_arr[19],x,y);};
 static double hdl20(x,y) double x,y; {return evaluate(tr_arr[20],x,y);};
 static double hdl21(x,y) double x,y; {return evaluate(tr_arr[21],x,y);};
 static double hdl22(x,y) double x,y; {return evaluate(tr_arr[22],x,y);};
 static double hdl23(x,y) double x,y; {return evaluate(tr_arr[23],x,y);};
 static double hdl24(x,y) double x,y; {return evaluate(tr_arr[24],x,y);};
 static double hdl25(x,y) double x,y; {return evaluate(tr_arr[25],x,y);};
 static double hdl26(x,y) double x,y; {return evaluate(tr_arr[26],x,y);};
 static double hdl27(x,y) double x,y; {return evaluate(tr_arr[27],x,y);};
 static double hdl28(x,y) double x,y; {return evaluate(tr_arr[28],x,y);};
 static double hdl29(x,y) double x,y; {return evaluate(tr_arr[29],x,y);};
 static double hdl30(x,y) double x,y; {return evaluate(tr_arr[30],x,y);};
 static double hdl31(x,y) double x,y; {return evaluate(tr_arr[31],x,y);}; 
#endif

static FP     hdl_arr[MAXHANDLE] ={ 
              hdl00, hdl01, hdl02, hdl03, hdl04, hdl05, hdl06, hdl07, 
              hdl08, hdl09, hdl10, hdl11, hdl12, hdl13, hdl14, hdl15, 
              hdl16, hdl17, hdl18, hdl19, hdl20, hdl21, hdl22, hdl23,
              hdl24, hdl25, hdl26, hdl27, hdl28, hdl29, hdl30, hdl31 };

#ifdef __cplusplus
  }        /* end for c++ */
#endif
