/****************************************************************
 *
 *    Copyright (c) 1993 1994 by Ke Jin, Queen's U. Canada
 *
 *    Permission to use, copy, modify, and distribute
 *    this software and its documentation without fee
 *    is granted, provided that the author's name and
 *    this copyright notice are retained.
 *
 *    This software can be included in other programs
 *    or software products, provided the source codes
 *    are also supplied without charge.
 *
 * ------------------------------------------------------------
 *
 *   dfcparse.y -- the yacc file of defunc parser 
 *
 *   public  function : yyparse();
 *                      yyinit();
 *                      getparsetree();
 *                      gettreename();
 *                      isfnctdecl();
 *                      dfcmin();
 *                      dfcmax();
 *
 *   private function : addnode();
 *                      yyreverse();
 *                      yyerror();
 *                      setrange();
 *
 *   private variable : yyparsetree;
 *                      yytreesize;
 *                      yytreename;
 *                      yydecltype;
 *                      exparserror;
 *                      range1;
 *                      range2;
 *                      range3;
 *                      newnode;
 *                    
 *
 ****************************************************************/

%{
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <string.h>
#include "dfctree.h"
#include "dfcsymtable.h"

typedef struct {
    double a; 
    double b;
    int    flag;
} Range;

static char*  exparserror;
static char   ermsgbuff[128];
static Node*  yyparsetree; 
static int    yytreesize;
static char*  yytreename;
static Symbol_type yydecltype;
static Node   newnode;
static Range  range[3];

#if NeedFunctionPrototypes
  int yyparse(void);
  static int yyreverse(void);
  static int yyerror(char* str);
  static int setrange(Range* ptr, int i);
#else
  extern int yyparse();
  extern int yyreverse();
  extern int yyerror();
  extern int setrange();
#endif

#if NeedFunctionPrototypes
  static int addnode(Node *ptr)
#else
  static int addnode(ptr)
  Node *ptr;
#endif
{
    yytreesize++;

    if(yytreesize==1)
    {
        yyparsetree = (Node*)malloc(sizeof(Node));
    }
    else
    {
        yyparsetree = (Node*)realloc((Node*)yyparsetree, 
                                 yytreesize*sizeof(Node));
    }

    if(yyparsetree==0)
    {
	perror("malloc/realloc in addnode()");
	exit(1);
    }

    /* yyparsetree[yytreesize-1] = *ptr; */
    memcpy(yyparsetree+yytreesize-1, ptr, sizeof(Node)); 

    return yytreesize-1;
};

%}

%union {
    int    nodeidx;
    int    argidx;
    double value;
    double (*fnctptr)();
    char   name[32];
    Range  range1d;
}

%token <value>   CONST  /* constant              */
%token <value>   DCONST /* constant for range    */
%token <argidx>  ARG    /* function arguments    */
%token <fnctptr> FNCT   /* intrinsic function    */
%token <name>    SYM    /* new symbol            */
%token <name>    DSYM   /* symbol before '='     */
%type  <nodeidx> expr   /* function expression   */
%type  <value>   cexpr  /* expression for range  */
%type  <name>    dkltor /* declarator            */ 
%type  <range1d> range

%left  '-' '+'
%left  '*' '/'
%left  SIG
%right '^'

%% /*  ------------------- syntax rules ------------------------ */ 
input : endmk             {
			      return yyerror("empty expression body");
                          }
      | expr endmk        {
			      return yyreverse();
                          }
      | ranges endmk      {
			      return yyerror("empty expression body");
                          }
      | ranges expr endmk {
			      return yyreverse();
                          }
      | spec '=' endmk    {
			      return yyerror("empty expression body");
                          }
      | spec '=' expr endmk  {
			      return yyreverse();
                          }
      | error             {
			      return -1;
                          }
      ;

endmk : '\n'|';'           /* end mark of an input */  
      ;

spec  : ranges 
      | decl 
      | ranges decl 
      ;

ranges: range             {
			      setrange(&$1, 1);
                          }
      | range range       {
			      setrange(&$1, 1);
			      setrange(&$2, 2);
                          }
      | range range range {
			      setrange(&$1, 1);
			      setrange(&$2, 2);
			      setrange(&$3, 3);
                          }
      ;

range : '[' ']'           { 
			      $$.flag = 0;
			  }                
      | '[' ',' ']'       { 
			      $$.flag = 0;
			  }
      | '[' cexpr ']'     {
			      $$.a = $2;
			      $$.flag = 1;
                          }
      | '[' cexpr ',' ']' { 
			      $$.a = $2;
			      $$.flag = 1;
                          }
      | '[' ',' cexpr ']' {
			      $$.b = $3;
			      $$.flag = 2;
                          }
      | '[' cexpr ',' cexpr ']'
			  {
			      $$.a = $2;
			      $$.b = $4;
			      $$.flag = 1+2;
                          }
      ;

cexpr : DCONST            {
			      $$ = $1;
                          }
      | FNCT '(' cexpr ')'{
			      $$ = (*$1)($3);
                          }
      | FNCT '(' cexpr ',' cexpr ')'
			  {
			      $$ = (*$1)($3,$5);
                          }
      | cexpr '+' cexpr   {
			      $$ = $1+$3;
                          }
      | cexpr '-' cexpr   {
			      $$ = $1-$3;
                          }
      | cexpr '*' cexpr   { 
			      $$ = $1*$3;
                          }
      | cexpr '/' cexpr   {
			      $$ = $1/$3;
                          }
      | '-' cexpr %prec SIG {
			      $$ = -$2;
                          }
      | '+' cexpr %prec SIG {
			      $$ =  $2;
                          }
      | cexpr '^' cexpr   {
			      $$ = pow($1, $3);
                          }
      | '(' cexpr ')'     {
			      $$ = $2;
                          }
      ;

decl  : dkltor            {
			      yytreename = $1;
			      yydecltype = const_symbol;
                          }
      | dkltor '(' ')'    {
			      yytreename = $1;
			      yydecltype = fnct_symbol;
                          }
      | dkltor '(' DSYM ')' {
			      yytreename = $1;
			      yydecltype = fnct_symbol;
			      if(strcmp($3, getarguname(1))!=0)
			      {
			          nameargu($3, "_");
                              }
                          }
      | dkltor '(' DSYM ',' ')'
			  {
			      yytreename = $1;
			      yydecltype = fnct_symbol;
			      if(strcmp($3, getarguname(1))!=0)
			      {
			          nameargu($3, "_");
                              }
                          }
      | dkltor '(' DSYM ',' DSYM ')' {
			      yytreename = $1;
			      yydecltype = fnct_symbol;
			      nameargu($3, $5);
                          }
      ;

dkltor: '$'               {
			      strcpy($$, "_");
                          }
      | '_'               {
			      strcpy($$, "_");
                          }
      | '$' DSYM          {
			      sprintf($$, "%c%s", '$', $2);
                          }
      | '_' DSYM          {
			      sprintf($$, "%c%s", '_', $2); 
                          }
      | DSYM              {
			      strncpy($$, $1, 32);;
                          }
      ;


expr  : CONST              {  
                              newnode.type = const_node;
                              newnode.content.value = $1;

                              $$ = addnode(&newnode); 
                           }
      | SYM                {
			      sprintf(ermsgbuff, 
				  "unknown token \"%s\"", $1);
			      return yyerror(ermsgbuff);
                           }
      | CONST '(' ')'      {
			      newnode.type = const_node;
			      newnode.content.value = $1;

			      $$ = addnode(&newnode);
                           }
      | CONST '(' expr ')' {
			      return
			      yyerror("constant be used as function");
                           }
      | ARG                {
                              newnode.type = arg_node;
			      newnode.content.argidx = $1;
                              $$ = addnode(&newnode);
                           }
      | ARG '('            {
			      return 
			      yyerror("argument used as function/constant");
                           }
      | FNCT               {
			      return
			      yyerror("use function without argument(s)");
                           }
      | FNCT '(' ')'       {
			      return
			      yyerror("use function without argument(s)");
                           }
      | FNCT '(' expr ')'  { 
                              newnode.type = simplex_fnct_node;
                              newnode.content.fnctptr = $1;
                              newnode.right = $3;

                              $$ = addnode(&newnode);
                           }
      | FNCT '(' expr ',' expr ')' {
                              newnode.type = duplex_fnct_node;
                              newnode.content.fnctptr = $1;
                              newnode.left = $3;
                              newnode.right= $5;

                              $$ = addnode(&newnode);
                           }
      | FNCT '(' expr ',' expr ',' expr ')' {
			      return
			      yyerror("not support triplex function yet");
	                   } 
      | expr '+' expr      {
                              newnode.type = binary_op_node;
                              newnode.content.op = op_sum;
                              newnode.left = $1;
                              newnode.right= $3;

                              $$ = addnode(&newnode);
                           }
      | expr '-' expr      {
                              newnode.type = binary_op_node;
                              newnode.content.op = op_sub;
                              newnode.left = $1;
                              newnode.right= $3;

                              $$ = addnode(&newnode);
                           }
      | expr '*' expr      {
                              newnode.type = binary_op_node;
                              newnode.content.op = op_mul;
                              newnode.left = $1;
                              newnode.right= $3;

                              $$ = addnode(&newnode);
                           }
      | expr '/' expr      {
                              newnode.type = binary_op_node;
                              newnode.content.op = op_div;
                              newnode.left =$1;
                              newnode.right=$3;

                              $$ = addnode(&newnode); 
                           }
      | '-' expr %prec SIG {
                              newnode.type = unary_op_node; 
			      newnode.content.op = op_neg;
                              newnode.right = $2;

                              $$ = addnode(&newnode);
                           }
      | '+' expr %prec SIG { 
                              $$ = $2;
                           } 
      | expr '^' expr      {
                              newnode.type = duplex_fnct_node;
                              newnode.content.fnctptr = pow;
                              newnode.left = $1;
                              newnode.right= $3;

                              $$ = addnode(&newnode);
                           } 
      | '(' expr ')'       { 
                              $$ = $2; 
                           }
      ;
%% /* --------------------------------------------------------- */
#include "dfcscan.h"

#if NeedFunctionPrototypes
  int yyinit(char* expr)
#else
  int yyinit(expr)
  char *expr; 
#endif
{
    initargu();

    yyexpr = expr;
    yyexprlen = strlen(yyexpr);
    yypos = 0;
    yytreesize = 0;
    yydecl = 0;
    yyrange = 0;
    exparserror = 0;
    range[1].flag = 0;
    range[2].flag = 0;
    range[3].flag = 0;

    return 0;
};

#if NeedFunctionPrototypes
  static int yyreverse(void)
#else
  static int yyreverse()  
#endif
/* yyparse() use a LALR(1) bottom-up algorithem to construct the
   parse tree. Thus the result tree is upsetdown, i.e. the root
   is place on the end of the yyparsetree[]. yyreverse make it
   in right order, i.e. yyparsetree[0] be the root */
{
    int   i;
    Node* buff;

    if(yytreesize==0) return 0;

    buff = (Node*)malloc(sizeof(Node)*yytreesize);
    if(buff==0) 
    {
	perror("malloc in reverse()");
	exit(1);
    }

    for(i=0; i<yytreesize; i++) /* reverse */
    {
	
 	/* buff[i] = yyparsetree[yytreesize-1-i]; */
	memcpy(buff+i, yyparsetree+yytreesize-1-i, sizeof(Node));
	buff[i].left = yytreesize - 1 - buff[i].left;
	buff[i].right= yytreesize - 1 - buff[i].right;

    }

    for(i=0; i<yytreesize; i++) /* put it back */ 
    {
	yyparsetree[i] = buff[i];
    }

    free(buff);

    return yytreesize;
};

#if NeedFunctionPrototypes
  int getparsetree(Node* buff)
#else
  int getparsetree(buff)
  Node* buff;
#endif
/* copy the parse into buff */
{
    int i;

    for(i=0; i<yytreesize; i++)
    {
	buff[i] = yyparsetree[i];
    }

    return yytreesize;
};

char* gettreename()
{
  return yytreename;
};

int isfnctdecl()
{
    return yydecltype==fnct_symbol;
};

#if NeedFunctionPrototypes
  double dfcmin(int i) 
#else
  double dfcmin(i)
  int i;
#endif
{ 
   double a, b;

   if(i<1||i>3) return 0.0;

   a = range[i-1].a;
   b = range[i-1].b;

   return (a<b)? a:b;
};

#if NeedFunctionPrototypes
  double dfcmax(int i)
#else
  double dfcmax(i)
  int i;
#endif
{
    double a, b;

    if(i<1||i>3) return 0.0;

    a = range[i-1].a;
    b = range[i-1].b;

    return (a<b)?  b:a;
};

#if NeedFunctionPrototypes
  static int setrange(Range* ptr, int i)
#else
  static int setrange(ptr, i) 
  Range* ptr;
  int i;
#endif
{
    if(i<1||i>3) return 0;

    switch(ptr->flag)
    {
	case 0: 
	     break;

	case 1: 
	     range[i-1].b = dfcmax(i);
	     range[i-1].a = ptr->a;
	     break;

	case 2:
	     range[i-1].a = dfcmin(i);
	     range[i-1].b = ptr->b;
	     break;

        case 3:
	     range[i-1].a = ptr->a;
	     range[i-1].b = ptr->b;
	     break;

        default: break;
    }

    return ptr->flag; 
};

#if NeedFunctionPrototypes
  static int yyerror(char* s)
#else
  static int yyerror(s)
  char *s;
#endif
{
    if (exparserror==0) exparserror=s;
    return -1;
};

char* dferror()
/* return the error on the last yyparse() call */
{ 
    return (exparserror==0)? "":exparserror;
};

