/****************************************************************
 *
 *    Copyright (c) 1993 1994 by Ke Jin, Queen's U. Canada.
 *
 *    Permission to use, copy, modify, and distribute
 *    this software and its documentation without fee
 *    is granted, provided that the author's name and
 *    this copyright notice are retained.
 *
 *    This software can be included in other programs
 *    or software products, provided the source codes
 *    are also supplied without charge.
 *
 * -------------------------------------------------------------
 *  
 *   dfcscan.h --- the lexical scaner of defunc
 *
 *   private variable : yyexpr;
 *                      yyexprlen;
 *                      yytext;
 *                      yypos;
 *
 *   private function : yylex();
 *
 ****************************************************************/

#ifndef _DFCSCAN_H
#define _DFCSCAN_H

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "dfcsymtable.h"

#ifdef __cplusplus
  extern "C" {    /* for c++ */
#endif

static char* yyexpr;
static int   yyexprlen;
static char* yytext; 
static int   yypos;
static int   yyrange;   /* in sider a [ ] ? */
static int   yydecl;    /* before a '='   ? */ 

#if NeedFunctionPrototypes
  static int yylex(void)
#else
  static int yylex()
#endif
{
    char c=' ';    
    int  i, j, k, eq;

    Symbol_record* ptr;

    /* -- context sensitive prescan ------------------------------ */
    if(yypos==0) 
    {
	for(i=0, j=0, k=0, eq=0;i<strlen(yyexpr);i++)
	{
	    c=yyexpr[i]; 
	    if(c==';'||c=='\n') break;  /* expressions will endup
					 * with ';', '\n' or '\0' */ 
	    switch(c)
	    {
		case '(': j++;  break;
		case ')': j--;  break;
		case '[': k++;  break;
		case ']': k--;  break;
		case '=': eq++; break;
		default:        break;
            }
	    if(j==-1)
	    {
		yyerror("Too many )'s.");
		return YYERRCODE;
            }
	    if(k==-1)
	    {
		yyerror("Too many ]'s.");
		return YYERRCODE; 
            }
	    if(k==2)      /* [] should not embedded */
	    {
		yyerror("Too many ['s.");
		return YYERRCODE;
            }
	    if(eq==2)     /* one expression should include  
			   * no more than one declaration. */ 
	    {
		yyerror("Too many ='s");
		return YYERRCODE;
            } 
        }
	if(j!=0)          /* ( ) should be balanced */
	{
	    yyerror("Too many ('s.");
	    return YYERRCODE;
        }
        yydecl = eq;      /* context sensitive flag */	
	c=' ';
    }

    /* --- [ \t] do nothing --------------------------------- */
    for(;yypos<yyexprlen;yypos++)
    {
        c=yyexpr[yypos];
        if(c==' '||c=='\t') continue;
        break; 
    }

    if(c=='[') yyrange = 1;     /* inside a [ ] */
    if(c==']') yyrange = 0;     /* outside  [ ] */
    if(c=='=') yydecl  = 0;     /* not before = */ 

    /* --- {number} return token CONST or DCONST with value --- */ 
    if(c=='.'||isdigit(c))
    {
        sscanf(yyexpr+yypos, "%lf", &(yylval.value));
        for(;yypos<=yyexprlen;yypos++)
        {
	    c=yyexpr[yypos];
	    if(c=='.'||isdigit(c)) continue;
	    break;
        }
        if(yyrange==1) return DCONST;   /* inside '[ ]'    */
        if(yyrange==0) return CONST;    /* outside '[' ']' */
	return YYERRCODE;
    }

    if(c=='*')
    {
	if(yyexpr[yypos+1]=='*')
	{
	    yypos=yypos+2;
	    c=yyexpr[yypos];
	    return '^';
        }
	else
	{
	    yypos++;
	    c=yyexpr[yypos];
	    return '*';
        }
    }

    /* --- {symbol} return token VAR or FNCT with fnctptr --- */
    if(isalpha(c))
    {
        yytext = (char*)malloc(sizeof(char)*(yyexprlen-yypos+1));
        if(yytext==0) 
        {
	    perror("malloc for token text in module scaner");
	    exit(1);
        }

        for(i=0;yypos<=yyexprlen;yypos++, i++)
        {
	    c=yyexpr[yypos];
	    if(isalnum(c))
	    {
	        yytext[i]=c;
	        continue;
            }
	    yytext[i]='\0';
	    break;
        }

	if(yydecl==1&&yyrange==0) 
	/* before '=' and outside '[ ]' */
	{
	    strncpy(yylval.name, yytext, 32);
	    return DSYM;
        }

        ptr = (Symbol_record*)getsym(yytext);
        if(ptr==0) /* symbol not in table */
        {
 	    strncpy(yylval.name, yytext, 32);
	    return SYM;
        }

        switch (ptr->type)
        {
	    case arg_symbol:
		 yylval.argidx = ptr->content.argidx;
		 if(yyrange==0) return ARG; 
		 if(yyrange==1)
		 {
		     yyerror("argument symbol in range specification");
 		     return YYERRCODE; 
                 }
		 return YYERRCODE;
 
	    case fnct_symbol:
	         yylval.fnctptr = ptr->content.fnctptr; 
	         return FNCT;

            case const_symbol:
	         yylval.value = ptr->content.value; 
		 if(yyrange==0) return CONST;
	         if(yyrange==1) return DCONST;
                 yyerror("argument symbol in range specification");
		 return YYERRCODE;

	    default: return YYERRCODE;
        }
    }

   /* ------ append a '\n' on end ----------------- */
   if(yypos==yyexprlen) return '\n';

   yypos++;
   /* ------ return other characters -------------- */
   return c;
};

#ifdef __cplusplus
  }    /* end for c++ */
#endif

#endif /* _DFCSCAN_H */
