/************************************************************
 *
 *    Copyright (c) 1993 1994 by Ke Jin, Queen's U. Canada.
 *
 *    Permission to use, copy, modify, and distribute
 *    this software and its documentation without fee
 *    is granted, provided that the author's name and
 *    this copyright notice are retained.
 *
 *    This software can be included in other programs
 *    or software products, provided the source codes
 *    are also supplied without charge.
 *
 * --------------------------------------------------------
 *
 *    dfctree.h -- interface of defunc low level module
 *
 *    struct definition : Node
 *
 *    external function : exparse();
 *                        dferror();
 *			  evaluate();
 *                        getparsetree(); 
 *                        gettreename();
 *                        getexprbody();
 *                        isfnctdecl();
 *                        dfcmin();
 *                        dfcmax();
 *                        dfcreduce();
 *
 ************************************************************/

#ifndef _DFCTREE_H
#define _DFCTREE_H

#ifndef NeedFunctionPrototypes
#if defined(__STDC__)||defined(__cplusplus)
#define NeedFunctionPrototypes 1
#else 
#define NeedFunctionPrototypes 0
#endif  /* __STDC__ */
#endif  /* NeedFunctionPrototypes */

#ifdef __cplusplus
  extern "C" {    /* for c++ */
#endif

typedef enum { 
    const_node, 
    arg_node, 

    unary_op_node,
    binary_op_node, 

    simplex_fnct_node, 
    duplex_fnct_node,
} Node_type;

typedef enum {
    op_sum,
    op_sub,
    op_mul,
    op_div,
    op_neg
} op_idx;

typedef struct {
    Node_type type;
    union {
       int    argidx;
       op_idx op;
       double value;
       double (*fnctptr)();
    } content;
    int left;
    int right;
} Node;

#if NeedFunctionPrototypes /* ANSI */
   extern int    exparse(char* expression);
   extern char*  dferror(void);
   extern double evaluate(Node* tree, double x, double y);
   extern int    getparsetree(Node* buff);
   extern Node*  dfcreduce(Node* tree);
   extern char*  gettreename(void);
   extern char*  getexprbody(char* expression);
   extern int    isfnctdecl(void);
   extern double dfcmin(int i);
   extern double dfcmax(int i);
#else /* K&R */
   extern int    exparse();
   extern char*  dferror();
   extern double evaluate();
   extern int    getparsetree();
   extern Node*  dfcreduce();
   extern char*  gettreename();
   extern char*  getexprbody();
   extern int    isfnctdecl();
   extern double dfcmin();
   extern double dfcmax();
#endif /* NeedFunctionPrototypes */

#ifdef __cplusplus
  }    /* end for c++ */
#endif

#endif /* _DFCTREE_H */
